/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdlib.h>
#include <string.h>

#include <sqlite3.h>

#include "common.h"
#include "task.h"
#include "query.h"
#include "storage.h"
#include "sqlite.h"

typedef struct tm_sqlite_query tm_sqlite_query;

struct tm_sqlite_query
{
    char *name_contains;
    unsigned long *pid_equals;

    tm_query **and;
    unsigned int and_len;
};

static void sqlite_query_free(void *);

static char *query_create_table_tasks = "CREATE TABLE IF NOT EXISTS tasks (" \
    "id INTEGER PRIMARY KEY ASC AUTOINCREMENT DEFAULT 1, " \
    "name TEXT NOT NULL"\
    ")";
static char *query_run_query_prefix = "SELECT id, name FROM tasks";

static char *query_add_task = "INSERT INTO tasks (name) VALUES (?)";
static char *query_modify_task_prefix = "UPDATE tasks SET";
static char *query_modify_task_postfix = "WHERE id = ?";
static char *query_remove_task = "DELETE FROM tasks WHERE id = ?";

static void sqlite_free(tm_storage *);
static int sqlite_run_query(tm_storage *, tm_query *, tm_task ***parr,
    unsigned int *plen);
static int sqlite_add_tasks(tm_storage *, tm_task **tasks, unsigned int len);
static int sqlite_modify_tasks(tm_storage *, tm_task **tasks, unsigned int len);
static int sqlite_remove_tasks(tm_storage *, tm_task **tasks, unsigned int len);

static void sqlite_free(tm_storage *this)
{
    sqlite3_close((sqlite3 *)this->data);
}

static int process_row(sqlite3 *db, sqlite3_stmt *stmt, int rc, tm_task ***parr,
    unsigned int *plen)
{
    tm_task *task = 0;
    tm_task **arr = 0;

    if (rc == SQLITE_ROW)
    {
        (*plen)++;
        arr = realloc(*parr, sizeof(tm_task *) * (*plen));
        if (0 == arr)
        {
            fprintf(stderr, "Cannot realloc\n");
            return 0;
        }

        tm_task_new(&task);
        tm_task_set_id(task, sqlite3_column_int(stmt, 0));
        tm_task_set_name(task, (char *)sqlite3_column_text(stmt, 1));
        tm_task_invalidate(task);

        arr[(*plen) - 1] = task;
        *parr = arr;
    }
    else if (rc != SQLITE_DONE)
    {
        fprintf(stderr, "Cannot query, step returned %d: %s\n",
            rc, sqlite3_errmsg(db));
        return 0;
    }

    return 1;
}

static void serialize_query(char **pqtext, unsigned int *pqlen, tm_query *query)
{
    tm_sqlite_query *data = (tm_sqlite_query *)query->data;
    if (0 != data->name_contains)
    {
        tm_compact_sprints(pqtext, pqlen, "name GLOB ?");
    }
    else if (0 != data->pid_equals)
    {
        tm_compact_sprints(pqtext, pqlen, "id = ?");
    }
    else if (0 != data->and)
    {
        tm_compact_sprints(pqtext, pqlen, "(");
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            if (i > 0)
            {
                tm_compact_sprints(pqtext, pqlen, " AND ");
            }
            serialize_query(pqtext, pqlen, data->and[i]);
        }
        tm_compact_sprints(pqtext, pqlen, ")");
    }
}

static void bind_args_to_stmt(sqlite3_stmt *stmt, unsigned int *pi, tm_query *query)
{
    tm_sqlite_query *data = (tm_sqlite_query *)query->data;
    if (0 != data->name_contains)
    {
        char *tmp = 0;
        unsigned int tmplen = 0;
        tm_compact_sprints(&tmp, &tmplen, "*");
        tm_compact_sprints(&tmp, &tmplen, data->name_contains);
        tm_compact_sprints(&tmp, &tmplen, "*");
        sqlite3_bind_text(stmt, ++(*pi), tmp, tmplen, free);
    }
    else if (0 != data->pid_equals)
    {
        sqlite3_bind_int64(stmt, ++(*pi), *(data->pid_equals));
    }
    else if (0 != data->and)
    {
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            bind_args_to_stmt(stmt, pi, data->and[i]);
        }
    }
}

static void build_stmt(sqlite3 *db, tm_query *query, sqlite3_stmt **pstmt)
{
    sqlite3_stmt *stmt = 0;
    char *qtext = strdup(query_run_query_prefix);
    unsigned int qlen = strlen(qtext);
    unsigned int i = 0;

    if (0 != query)
    {
        tm_compact_sprints(&qtext, &qlen, " WHERE ");
        serialize_query(&qtext, &qlen, query);
    }

    sqlite3_prepare_v2(db, qtext, strlen(qtext), &stmt, 0);
    if (0 != query)
    {
        bind_args_to_stmt(stmt, &i, query);
    }
    free(qtext);
    *pstmt = stmt;
}

static int sqlite_run_query(tm_storage *this, tm_query *query, tm_task ***parr,
    unsigned int *plen)
{
    sqlite3_stmt *stmt = 0;
    tm_task **arr = 0;
    int rc;
    unsigned int len = 0;

    build_stmt((sqlite3 *)this->data, query, &stmt);
    arr = (tm_task **)malloc(sizeof(tm_task *) * 0);

    do
    {
        rc = sqlite3_step(stmt);
        if (0 == process_row((sqlite3 *)this->data, stmt, rc, &arr, &len))
        {
            free(arr);
            sqlite3_finalize(stmt);
            return 0;
        }
    } while (SQLITE_DONE != rc);
    sqlite3_finalize(stmt);
    *parr = arr;
    *plen = len;
    return 1;
}

static int begin_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "BEGIN TRANSACTION", 0, 0, 0);
}

static int end_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "END TRANSACTION", 0, 0, 0);
}

static int rollback_transaction(sqlite3 *db)
{
    return SQLITE_OK == sqlite3_exec(db, "ROLLBACK TRANSACTION", 0, 0, 0);
}


static int sqlite_add_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot add tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        sqlite3_stmt *stmt = 0;
        int rc;
        char *name = 0;

        tm_task_get_name(task, &name);
        sqlite3_prepare_v2(db, query_add_task, strlen(query_add_task),
            &stmt, 0);
        sqlite3_bind_text(stmt, 1, strdup(name), strlen(name), free);
        rc = sqlite3_step(stmt);
        sqlite3_finalize(stmt);
        if (SQLITE_DONE != rc)
        {
            rollback_transaction(db);
            fprintf(stderr, "Cannot add task %u/%u: %s\n", i, len, sqlite3_errmsg(db));
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot add tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task_invalidate(tasks[i]);
    }
    return 1;
}

static int modify_task(sqlite3 *db, tm_task *task)
{
    sqlite3_stmt *stmt = 0;
    char *name = 0;
    char *qtext = 0;
    unsigned int qlen = 0;
    unsigned int j = 0;
    int rc;

    tm_compact_sprints(&qtext, &qlen, query_modify_task_prefix);
    if (0 != task->did_name_change)
    {
        tm_compact_sprints(&qtext, &qlen, " name = ?");
    }
    tm_compact_sprints(&qtext, &qlen, query_modify_task_postfix);

    tm_task_get_name(task, &name);
    sqlite3_prepare_v2(db, qtext, qlen, &stmt, 0);
    free(qtext);
    if (0 != task->did_name_change)
    {
        sqlite3_bind_text(stmt, ++j, strdup(task->name), strlen(task->name), free);
    }
    sqlite3_bind_int64(stmt, ++j, task->id);
    rc = sqlite3_step(stmt);
    sqlite3_finalize(stmt);
    return SQLITE_DONE == rc;
}

static int sqlite_modify_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i = 0;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot modify tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        if (0 == modify_task(db, task))
        {
            rollback_transaction(db);
            fprintf(stderr, "Cannot modify task %u/%u: %s\n", i + 1, len, sqlite3_errmsg(db));
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot modify tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task_invalidate(tasks[i]);
    }
    return 1;
}

static int sqlite_remove_tasks(tm_storage *this, tm_task **tasks, unsigned int len)
{
    sqlite3 *db = (sqlite3 *)this->data;
    unsigned int i;

    if (0 == begin_transaction(db))
    {
        fprintf(stderr, "Cannot remove tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    for (i = 0; i < len; i++)
    {
        tm_task *task = tasks[i];
        sqlite3_stmt *stmt = 0;
        int rc;
        unsigned long id = -1;

        tm_task_get_id(task, &id);
        sqlite3_prepare_v2((sqlite3 *)this->data, query_remove_task, strlen(query_remove_task),
            &stmt, 0);
        sqlite3_bind_int64(stmt, 1, id);
        rc = sqlite3_step(stmt);
        sqlite3_finalize(stmt);
        if (SQLITE_DONE != rc)
        {
            rollback_transaction(db);
            fprintf(stderr, "Cannot remove task %u/%u: %s\n", i, len, sqlite3_errmsg(db));
            return 0;
        }
    }

    if (0 == end_transaction(db))
    {
        rollback_transaction(db);
        fprintf(stderr, "Cannot remove tasks: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    return 1;
}

static void sqlite_query_free(void *dat)
{
    tm_sqlite_query *data = (tm_sqlite_query *)dat;
    if (0 != data->name_contains)
    {
        free(data->name_contains);
    }
    if (0 != data->pid_equals)
    {
        free(data->pid_equals);
    }
    if (0 != data->and)
    {
        for (unsigned int i = 0; i < data->and_len; i++)
        {
            tm_query_free(&(data->and[i]));
        }
        free(data->and);
    }
    free(data);
}

static int sqlite_query_name_contains(tm_storage *this, char *name, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->name_contains = strdup(name);
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_id_equals(tm_storage *this, unsigned long id, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->pid_equals = (unsigned long *)malloc(sizeof(unsigned long));
    *(data->pid_equals) = id;
    return tm_query_new(data, sqlite_query_free, pquery);
}

static int sqlite_query_and(tm_storage *this, unsigned int clen, tm_query **cs, tm_query **pquery)
{
    tm_sqlite_query *data = (tm_sqlite_query *)malloc(sizeof(tm_sqlite_query));
    memset(data, 0, sizeof(tm_sqlite_query));
    data->and = cs;
    data->and_len = clen;
    return tm_query_new(data, sqlite_query_free, pquery);
}

int tm_storage_sqlite(char *file, tm_storage **pthis)
{
    tm_storage *this = 0;
    sqlite3 *db = 0;
    int rc;
    char *errmsg = 0;

    if (0 == file || 0 == pthis || 0 != *pthis)
    {
        return 0;
    }

    if (0 != (rc = sqlite3_open(file, &db)))
    {
        fprintf(stderr, "Cannot open database: %s\n", sqlite3_errmsg(db));
        return 0;
    }

    if (0 != (rc = sqlite3_exec(db, query_create_table_tasks, 0, 0, &errmsg)))
    {
        sqlite3_close(db);
        fprintf(stderr, "Cannot create tables: %s\n", errmsg);
        return 0;
    }

    this = (tm_storage *)malloc(sizeof(struct tm_storage));
    this->data = (void *)db;
    this->free_f = sqlite_free;
    this->run_query_f = sqlite_run_query;
    this->add_tasks_f = sqlite_add_tasks;
    this->modify_tasks_f = sqlite_modify_tasks;
    this->remove_tasks_f = sqlite_remove_tasks;
    this->query_name_contains_f = sqlite_query_name_contains;
    this->query_id_equals_f = sqlite_query_id_equals;
    this->query_and_f = sqlite_query_and;
    *pthis = this;
    return 1;
}
