/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <criterion/criterion.h>

#include <tm.h>

static char filename[80];
static tm *this;

static void setup(void)
{
    tm_storage *storage = 0;
    sprintf(filename, "tmp.db.XXXXXX");
    mktemp(filename);
    cr_expect_eq(1, tm_storage_sqlite(filename, &storage));
    cr_expect_eq(1, tm_new(storage, &this));
}

static void teardown(void)
{
    tm_free(&this);
    remove(filename);
}

TestSuite(tm_empty, .init = setup, .fini = teardown);

Test(tm_empty, get_all_tasks)
{
    tm_task **arr = 0;
    unsigned int len = -1;

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(0, len);

    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);
}

Test(tm_empty, add_task)
{
    tm_task *task = 0;
    tm_task *tasks[1];
    tm_task **arr = 0;
    tm_query *query = 0;
    unsigned int len = -1;
    unsigned long id = -1;
    char *name = 0;

    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    tasks[0] = task;
    cr_expect_eq(1, tm_add_tasks(this, tasks, 1));
    cr_expect_eq(1, tm_task_free(&task));

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(1, len);
    cr_expect_neq(0, arr[0]);
    cr_expect_eq(1, tm_task_get_id(arr[0], &id));
    cr_expect_eq(1, id);
    cr_expect_eq(1, tm_task_get_name(arr[0], &name));
    cr_expect_str_eq("Buy milk", name);
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);

    cr_expect_eq(1, tm_query_name_contains(this, "milk", &query));
    cr_expect_neq(0, query);
    cr_expect_eq(1, tm_run_query(this, query, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(1, len);
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);
    cr_expect_eq(1, tm_query_free(&query));
    cr_expect_eq(0, query);
}

Test(tm_empty, add_modify_task)
{
    tm_task *task = 0;
    tm_task *tasks[1];
    tm_task **arr = 0;
    unsigned int len = -1;
    unsigned long id = -1;
    char *name = 0;

    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    tasks[0] = task;
    cr_expect_eq(1, tm_add_tasks(this, tasks, 1));
    cr_expect_eq(1, tm_task_free(&task));

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(1, len);
    cr_expect_neq(0, arr[0]);
    cr_expect_eq(1, tm_task_set_name(arr[0], "Buy beer"));
    cr_expect_eq(1, tm_modify_tasks(this, arr, len));
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(1, len);
    cr_expect_neq(0, arr[0]);
    cr_expect_eq(1, tm_task_get_id(arr[0], &id));
    cr_expect_eq(1, id);
    cr_expect_eq(1, tm_task_get_name(arr[0], &name));
    cr_expect_str_eq("Buy beer", name);
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);
}

Test(tm_empty, add_modify_tasks)
{
    tm_task **tasks = 0;
    tm_task **arr = 0;
    unsigned int len = -1;
    unsigned int i;

    tasks = (tm_task **)malloc(sizeof(tm_task *) * 3);
    memset(tasks, 0, sizeof(tm_task *) * 3);
    for (i = 0; i < 3; i++)
    {
        cr_expect_eq(1, tm_task_new(&(tasks[i])));
        cr_expect_eq(1, tm_task_set_name(tasks[i], "Buy milk"));
    }
    cr_expect_eq(1, tm_add_tasks(this, tasks, 3));
    cr_expect_eq(1, tm_task_arr_free(&tasks, 3));

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(3, len);
    cr_expect_eq(1, tm_task_arr_set_name(arr, len, "Buy beer"));
    cr_expect_eq(1, tm_modify_tasks(this, arr, len));
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(3, len);
    for (i = 0; i < 3; i++)
    {
        char *name = 0;
        cr_expect_eq(1, tm_task_get_name(arr[i], &name));
        cr_expect_str_eq("Buy beer", name);
    }
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);
}

Test(tm_empty, add_remove_task)
{
    tm_task *task = 0;
    tm_task *tasks[1];
    tm_task **arr = 0;
    unsigned int len = -1;

    cr_expect_eq(1, tm_task_new(&task));
    cr_expect_eq(1, tm_task_set_name(task, "Buy milk"));
    tasks[0] = task;
    cr_expect_eq(1, tm_add_tasks(this, tasks, 1));
    cr_expect_eq(1, tm_task_free(&task));

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(1, len);
    cr_expect_eq(1, tm_remove_tasks(this, arr, 1));
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);

    cr_expect_eq(1, tm_run_query(this, 0, &arr, &len));
    cr_expect_neq(0, arr);
    cr_expect_eq(0, len);
    cr_expect_eq(1, tm_task_arr_free(&arr, len));
    cr_expect_eq(0, arr);
}
