/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

const request = require('browser-request');

const doQuery = (opts) => {
  return new Promise((resolve, reject) => {
    let args = {
      url: '/query',
      headers: {'Content-Type': 'application/json'},
    };
    if (opts !== undefined && Object.keys(opts).length > 0) {
      args.method = 'POST';
      args.body = JSON.stringify(opts);
    }
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        try {
          resolve(JSON.parse(res.body));
        } catch (e) {
          reject(e);
        }
      }
    });
  });
};

const doAdd = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject('opts must be non-empty object');
      return;
    }
    const args = {
      url: '/add',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const doRemove = (opts) => {
  return new Promise((resolve, reject) => {
    if (opts === undefined || Object.keys(opts).length == 0) {
      reject('opts must be non-empty object');
      return;
    }
    let args = {
      url: '/remove',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify(opts),
    };
    request(args, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const doModify = (query, args) => {
  return new Promise((resolve, reject) => {
    if (query === undefined || Object.keys(query).length == 0) {
      reject('query must be non-empty object');
      return;
    }
    if (args === undefined || Object.keys(args).length == 0) {
      reject('args must be non-empty object');
      return;
    }
    let requestArgs = {
      url: '/modify',
      headers: {'Content-Type': 'application/json'},
      method: 'POST',
      body: JSON.stringify({'query': query, 'args': args}),
    };
    request(requestArgs, (err, res) => {
      if (err) {
        reject(err);
      } else {
        resolve();
      }
    });
  });
};

const addColumnTr = (table, columns) => {
  const tr = document.createElement('tr');
  table.appendChild(tr);

  /* For remove */
  tr.appendChild(document.createElement('td'));
  /* For modify */
  tr.appendChild(document.createElement('td'));

  for (let i = 0; i < columns.length; i++) {
    const td = document.createElement('td');
    tr.appendChild(td);

    const column = columns[i];
    td.appendChild(document.createTextNode(column));
  }
};

const addRemoveButton = (tr, itemId) => {
  const td = document.createElement('td');
  tr.appendChild(td);

  const button = document.createElement('button');
  td.appendChild(button);
  button.appendChild(document.createTextNode('Remove'));
  button.onclick = () => {
    window.uiDoRemove(itemId, removeTd);
  };
};

const addModifyDiv = (tr, itemId) => {
  const td = document.createElement('td');
  tr.appendChild(td);

  const div = document.createElement('div');
  td.appendChild(div);

  const pButton = document.createElement('p');
  div.appendChild(pButton);

  const button = document.createElement('button');
  pButton.appendChild(button);
  button.appendChild(document.createTextNode('Modify'));

  const pName = document.createElement('p');
  div.appendChild(pName);

  pName.appendChild(document.createTextNode('Name:'));

  const inputName = document.createElement('input');
  pName.appendChild(inputName);
  inputName.setAttribute('id', 'modify-' + itemId + '-name');

  button.onclick = () => {
    const name = inputName.value;
    window.uiDoModify(itemId, pButton, name);
  };
};

const addItemTr = (table, columns, item) => {
  const tr = document.createElement('tr');
  table.appendChild(tr);

  addRemoveButton(tr, item['id']);
  addModifyDiv(tr, item['id']);

  for (let i = 0; i < columns.length; i++) {
    const td = document.createElement('td');
    tr.appendChild(td);

    const column = columns[i];
    let value = '-';
    if (item.hasOwnProperty(column)) {
      value = item[column];
    }
    td.appendChild(document.createTextNode(value));
  }
};

const uiDoQuery = () => {
  return new Promise((resolve) => {
    const idEquals = document.getElementById('query-id-equals').value;
    const nameContains = document.getElementById('query-name-contains').value;
    const queryResultDiv = document.getElementById('query-result');
    const resultsLenDiv = document.getElementById('results-len');
    const resultsDiv = document.getElementById('results');
    let queryArgs = {};

    queryResultDiv.innerHTML = '';
    resultsLenDiv.innerHTML = '';
    resultsDiv.innerHTML = '';

    if (idEquals !== undefined && idEquals !== '') {
      queryArgs['id-equals'] = parseInt(idEquals);
      if (isNaN(queryArgs['id-equals'])) {
        queryResultDiv.appendChild(document.createTextNode('ID must be an integer!'));
        resolve();
        return;
      }
    }
    if (nameContains !== undefined && nameContains !== '') {
      queryArgs['name-contains'] = nameContains;
    }

    doQuery(queryArgs).then((ret) => {
      resultsLenDiv.appendChild(document.createTextNode('Results: ' + ret.length));
      queryResultDiv.appendChild(document.createTextNode('OK'));

      let columns = {};
      for (let i = 0; i < ret.length; i++) {
        const item = ret[i];
        const itemColumns = Object.keys(item);
        for (let j = 0; j < itemColumns.length; j++) {
          const column = itemColumns[j];
          columns[column] = 0;
        }
      }
      columns = Object.keys(columns);

      const table = document.createElement('table');
      resultsDiv.appendChild(table);

      if (columns.length > 0) {
        addColumnTr(table, columns);
      }
      for (let i = 0; i < ret.length; i++) {
        const item = ret[i];
        addItemTr(table, columns, item);
      }
      resolve();
    }, (err) => {
      queryResultDiv.appendChild(document.createTextNode('ERR ' + err));
      resolve();
    });
  });
};

const uiDoAdd = () => {
  return new Promise((resolve) => {
    const name = document.getElementById('add-name').value;
    const addResultDiv = document.getElementById('add-result');
    let queryArgs = {};

    addResultDiv.innerHTML = '';
    if (name !== undefined && name !== '') {
      queryArgs['name'] = name;
    }

    doAdd(queryArgs).then(() => {
      addResultDiv.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      addResultDiv.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

const uiDoRemove = (anId, anTd) => {
  return new Promise((resolve) => {
    let queryArgs = {'id-equals': anId};

    doRemove(queryArgs).then(() => {
      anTd.innerHTML = '';
      anTd.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      const button = anTd.children[0];
      anTd.innerHTML = '';
      anTd.appendChild(button);
      anTd.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

const uiDoModify = (anId, anTd, name) => {
  return new Promise((resolve) => {
    let queryArgs = {'id-equals': anId};
    let newArgs = {'name': name};

    doModify(queryArgs, newArgs).then(() => {
      anTd.innerHTML = '';
      anTd.appendChild(document.createTextNode('OK'));
      resolve();
    }, (err) => {
      const button = anTd.children[0];
      anTd.innerHTML = '';
      anTd.appendChild(button);
      anTd.appendChild(document.createTextNode('ERROR ' + err));
      resolve();
    });
  });
};

module.exports = {
  doAdd: doAdd,
  doQuery: doQuery,
  doRemove: doRemove,
  doModify: doModify,
  uiDoQuery: uiDoQuery,
  uiDoAdd: uiDoAdd,
  uiDoRemove: uiDoRemove,
};

if (window !== undefined) {
  window.uiDoQuery = uiDoQuery;
  window.uiDoAdd = uiDoAdd;
  window.uiDoRemove = uiDoRemove;
  window.uiDoModify = uiDoModify;
}
