/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>

#include <tm.h>

typedef struct argz argz;
struct argz
{
    int ca;
    int cd;
    int cm;
    int cr;
    int cq;

    char *d;
    int h;
    unsigned long *i;
    char *n;
    char *value;
};

static int parse_args(int, char **, argz *);
static void argz_free(argz *);

static void usage(char *);
static int do_add(argz *);
static int do_modify(argz *);
static int do_remove(argz *);
static int do_query(argz *);

static void usage(char *argv0)
{
    fprintf(stderr, "Usage: %s [-h]\n", argv0);
    fprintf(stderr, "       %s [opts] a{dd} [name]\n", argv0);
    fprintf(stderr, "       %s [opts] q{uery} [filters]\n", argv0);
    fprintf(stderr, "       %s [opts] r{emove} [filters]\n", argv0);
    fprintf(stderr, "       %s [opts] m{odify} [filters] [name]\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "       -d/--dbfile dbfile. Default: tm.db\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "Filters:\n");
    fprintf(stderr, "       -i/--id id\n");
    fprintf(stderr, "       -n/--name name\n");
    fprintf(stderr, "\n");
}

static int parse_command(char *command, int *phas_opts, int *phas_value,
    int *pmust_have_value, argz *arg)
{
    if (0 == strcmp("q", command) || 0 == strcmp("query", command))
    {
        *phas_opts = 1;
        arg->cq = 1;
    }
    else if (0 == strcmp("r", command) || 0 == strcmp("remove", command))
    {
        *phas_opts = 1;
        arg->cr = 1;
    }
    else if (0 == strcmp("a", command) || 0 == strcmp("add", command))
    {
        *phas_value = 1;
        *pmust_have_value = 1;
        arg->ca = 1;
    }
    else if (0 == strcmp("m", command) || 0 == strcmp("modify", command))
    {
        *phas_opts = 1;
        *phas_value = 1;
        *pmust_have_value = 1;
        arg->cm = 1;
    }
    else
    {
        return 0;
    }
    return 1;
}

static int parse_preopt_key(unsigned int *pi, char *k, char *pl)
{
    if (0 == strcmp("-d", k) ||
        0 == strcmp("--dbfile", k))
    {
        *pl = 'd';
    }
    else
    {
        return 0;
    }
    return 1;
}

static int parse_opt_key(int must_have_value, unsigned int *pi, char *k, char *pl)
{
    if (0 == strcmp("-i", k) ||
        0 == strcmp("--id", k))
    {
        *pl = 'i';
    }
    else if (0 == strcmp("-n", k) ||
        0 == strcmp("--name", k))
    {
        *pl = 'n';
    }
    else if (1 == must_have_value && 0 == strcmp("--", k))
    {
        (*pi)++;
        return 1;
    }
    else if (k[0] == '-')
    {
        return 0;
    }
    return 1;
}

static void parse_preopt_value(char k, char *v, argz *arg)
{
    if ('d' == k)
    {
        free(arg->d);
        arg->d = strdup(v);
    }
}

static void parse_opt_value(char k, char *v, argz *arg)
{
    if ('i' == k)
    {
        arg->i = (unsigned long *)malloc(sizeof(unsigned long));
        *(arg->i) = atol(v);
    }
    else if ('n' == k)
    {
        arg->n = v;
    }
}

static int parse_preopts(unsigned int *pi, int argc, char **argv, argz *arg)
{
    char k = 0;
    for (; (*pi) < argc; (*pi)++)
    {
        char *argvi = argv[*pi];
        if (0 != k)
        {
            parse_preopt_value(k, argvi, arg);
            k = 0;
        }
        else if (0 == parse_preopt_key(pi, argvi, &k))
        {
            break;
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_opts(int must_have_value, unsigned int *pi, int argc, char **argv, argz *arg)
{
    char k = 0;
    for (; (*pi) < argc; (*pi)++)
    {
        char *argvi = argv[*pi];
        if (0 == k)
        {
            if (0 == parse_opt_key(must_have_value, pi, argvi, &k))
            {
                return 0;
            }
            else if (0 == k)
            {
                break;
            }
        }
        else
        {
            parse_opt_value(k, argvi, arg);
            k = 0;
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_args(int argc, char **argv, argz *arg)
{
    unsigned int i = 1;
    int has_opts = 0;
    int has_value = 0;
    int must_have_value = 0;

    memset(arg, 0, sizeof(argz));
    arg->d = strdup("tm.db");

    if (argc == 1)
    {
        return 0;
    }

    if (0 == strcmp("-h", argv[1]) || 0 == strcmp("--help", argv[1]))
    {
        arg->h = 1;
        return 1;
    }

    if (0 == parse_preopts(&i, argc, argv, arg))
    {
        return 0;
    }

    if (0 == parse_command(argv[i++], &has_opts, &has_value, &must_have_value, arg))
    {
        return 0;
    }

    if (1 == has_opts && 0 == parse_opts(must_have_value, &i, argc, argv, arg))
    {
        return 0;
    }

    if (1 == has_value)
    {
        if (0 == argv[i] && 1 == must_have_value)
        {
            return 0;
        }
        arg->value = 0 != argv[i] ? argv[i] : "";
    }

    return 1;
}

static int do_add(argz *arg)
{
    tm *tm = 0;
    tm_storage *storage = 0;
    int ret = 0;
    tm_task *task = 0;
    tm_task *tasks[1];

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    tm_task_new(&task);
    tm_task_set_name(task, arg->value);

    tasks[0] = task;
    if (0 == tm_add_tasks(tm, tasks, 1))
    {
        ret = 1;
        fprintf(stderr, "Cannot add task\n");
    }
    tm_task_free(&task);
    tm_free(&tm);
    return ret;
}

static void build_query(argz *arg, tm *tm, tm_query **pquery)
{
    unsigned int children_len = 0;

    children_len = (0 != arg->i ? 1 : 0) + (0 != arg->n ? 1 : 0);
    if (0 < children_len)
    {
        tm_query **children = 0;
        unsigned int i = 0;

        children = (tm_query **)malloc(sizeof(tm_query *) * children_len);
        memset(children, 0, sizeof(tm_query *) * children_len);

        if (0 != arg->i)
        {
            tm_query_id_equals(tm, *(arg->i), &(children[i++]));
        }
        if (0 != arg->n)
        {
            tm_query_name_contains(tm, arg->n, &(children[i++]));
        }

        tm_query_and(tm, children_len, children, pquery);
    }
}

static int alog10(unsigned long a)
{
    unsigned int i = 0;
    for (; a > 0; i++, a/=10)
    {
    }
    return i;
}

static unsigned int get_id_column_len(tm_task **arr, unsigned int arr_len)
{
    unsigned int id_column_len = 2;
    for (unsigned i = 0; i < arr_len; i++)
    {
        unsigned long id = -1;
        unsigned int id_len;

        tm_task_get_id(arr[i], &id);
        id_len = alog10(id);
        if (id_len > id_column_len)
        {
            id_column_len = id_len;
        }
    }
    return id_column_len;
}

static int do_query(argz *arg)
{
    tm *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    unsigned int id_column_len;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    tm_run_query(tm, query, &arr, &arr_len);

    id_column_len = get_id_column_len(arr, arr_len);

    printf("ID%*s | Name\n", id_column_len - 2, "");
    for (unsigned i = 0; i < arr_len; i++)
    {
        unsigned long id = -1;
        char *name = 0;
        unsigned int id_len;

        tm_task_get_id(arr[i], &id);
        tm_task_get_name(arr[i], &name);

        id_len = alog10(id);
        printf("%lu%*s | %s\n", id, id_column_len - id_len, "", name);
    }
    tm_task_arr_free(&arr, arr_len);
    tm_query_free(&query);
    tm_free(&tm);
    return 0;
}

static int do_remove(argz *arg)
{
    tm *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    int ret = 1;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    if (0 == query)
    {
        fprintf(stderr, "Please provide at least one filter\n");
    }
    else
    {
        tm_run_query(tm, query, &arr, &arr_len);
        if (0 == arr_len)
        {
            fprintf(stderr, "No tasks to remove\n");
        }
        else if (0 == tm_remove_tasks(tm, arr, arr_len))
        {
            fprintf(stderr, "Cannot remove %u tasks\n", arr_len);
        }
        else
        {
            ret = 0;
        }
        tm_task_arr_free(&arr, arr_len);
        tm_query_free(&query);
    }
    tm_free(&tm);
    return ret;
}

static int do_modify(argz *arg)
{
    tm *tm = 0;
    tm_storage *storage = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    tm_query *query = 0;
    int ret = 1;

    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &tm);

    build_query(arg, tm, &query);
    if (0 == query)
    {
        fprintf(stderr, "Please provide at least one filter\n");
    }
    else
    {
        tm_run_query(tm, query, &arr, &arr_len);
        if (0 == arr_len)
        {
            fprintf(stderr, "No tasks to modify\n");
        }
        else
        {
            tm_task_arr_set_name(arr, arr_len, arg->value);
            if (0 == tm_modify_tasks(tm, arr, arr_len))
            {
                fprintf(stderr, "Cannot modify %u tasks\n", arr_len);
            }
            else
            {
                ret = 0;
            }
        }
        tm_task_arr_free(&arr, arr_len);
        tm_query_free(&query);
    }
    tm_free(&tm);
    return ret;
}

static void argz_free(argz *argz)
{
    if (0 != argz->i)
    {
        free(argz->i);
    }
    if (0 != argz->d)
    {
        free(argz->d);
    }
}

int main(int argc, char **argv)
{
    argz arg;
    int ret = 1;

    if (0 == parse_args(argc, argv, &arg))
    {
        usage(argv[0]);
    }
    else if (1 == arg.h)
    {
        usage(argv[0]);
        ret = 0;
    }
    else if (1 == arg.ca)
    {
        ret = do_add(&arg);
    }
    else if (1 == arg.cr)
    {
        ret = do_remove(&arg);
    }
    else if (1 == arg.cm)
    {
        ret = do_modify(&arg);
    }
    else if (1 == arg.cq)
    {
        ret = do_query(&arg);
    }
    argz_free(&arg);
    return ret;
}
