/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <criterion/criterion.h>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "common.h"

/* LCOV_EXCL_START */

static void fill_out(char **pout, unsigned int *plen, int pipe)
{
    char buf[4096];
    while (1)
    {
        int count = read(pipe, buf, sizeof(buf));
        char *tmp = 0;
        if (count == -1)
        {
            if (EINTR == errno)
            {
                continue;
            }
            cr_expect_fail("Cannot read() pout");
            break;
        }
        else if (count == 0)
        {
            break;
        }

        tmp = (char *)realloc(*pout, *plen + count);
        if (0 == tmp)
        {
            cr_expect_fail("Cannot realloc() pout");
            break;
        }
        memcpy(tmp + *plen, buf, count);
        *plen += count;
        *pout = tmp;
    }
}

static void verify_out(unsigned int expected_len, char *expected,
    unsigned int actual_len, char *actual)
{
    if (expected_len != actual_len || 0 != strncmp(expected, actual, expected_len))
    {
        cr_expect_fail("Outputs different\n"
            "<expected %u>\n%.*s\n</expected>\n<actual %u>\n%.*s\n</actual>",
        expected_len, expected_len, expected, actual_len, actual_len, actual);
    }
    if (0 != actual)
    {
        free(actual);
    }
}

void run_with_args(char **argv, unsigned int expected_exit, char *expected_stdout,
    unsigned int expected_stdout_len, char *expected_stderr, unsigned int expected_stderr_len)
{
    pid_t pid = 0;
    int stdout[2];
    int stderr[2];
    int status = 0;
    int actual_exit = -1;
    char *actual_stdout = 0;
    char *actual_stderr = 0;
    unsigned int actual_stdout_len = 0;
    unsigned int actual_stderr_len = 0;
    unsigned int i;

    cr_log_info("# Run with args:");
    for (i = 0; 0 != argv[i]; i++)
    {
        cr_log_info("    argv[%u] = %s", i, argv[i]);
    }
    cr_log_info("    expected_exit = %d", expected_exit);
    cr_log_info("    expected_stdout_len = %u", expected_stdout_len);
    cr_log_info("    <expected_stdout>\n%.*s\n    </expected_stdout>", expected_stdout_len,
        expected_stdout);
    cr_log_info("    expected_stderr_len = %u", expected_stderr_len);
    cr_log_info("    <expected_stderr>\n%.*s\n    </expected_stderr>", expected_stderr_len,
        expected_stderr);

    if (-1 == pipe(stdout))
    {
        cr_expect_fail("Cannot pipe() stdout");
        return;
    }
    if (-1 == pipe(stderr))
    {
        cr_expect_fail("Cannot pipe() stderr");
        return;
    }

    pid = fork();

    if (pid == -1)
    {
        cr_expect_fail("Cannot fork()");
        return;
    }

    if (pid == 0)
    {
        dup2(stdout[1], 1);
        dup2(stderr[1], 2);
        close(stdout[1]);
        close(stderr[1]);
        execv(argv[0], argv);
        cr_expect_fail("Cannot execv()");
        _exit(1);
    }

    close(stdout[1]);
    close(stderr[1]);

    fill_out(&actual_stdout, &actual_stdout_len, stdout[0]);
    verify_out(expected_stdout_len, expected_stdout,
        actual_stdout_len, actual_stdout);

    fill_out(&actual_stderr, &actual_stderr_len, stderr[0]);
    verify_out(expected_stderr_len, expected_stderr,
        actual_stderr_len, actual_stderr);

    close(stdout[0]);
    close(stderr[0]);
    waitpid(pid, &status, 0);
    actual_exit = WEXITSTATUS(status);
    cr_expect_eq(expected_exit, actual_exit, "Exit codes different, expected %d, actual %d",
        expected_exit, actual_exit);
}

void do_add_dbfile(char *dbfile, unsigned int len, char *arg)
{
    char *argv[6];
    unsigned int i = 0;

    argv[i++] = "../tm";
    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "a";
    argv[i++] = arg;
    argv[i++] = 0;

    run_with_args(argv, 0, "", 0, "", 0);
}

void do_add(unsigned int len, char *arg)
{
    do_add_dbfile(0, len, arg);
}

void do_remove(int found, char *id)
{
    char *argv[5];

    argv[0] = "../tm";
    argv[1] = "r";
    argv[2] = "-i";
    argv[3] = id;
    argv[4] = 0;

    if (1 == found)
    {
        run_with_args(argv, 0, "", 0, "", 0);
    }
    else
    {
        run_with_args(argv, 1, "", 0, "No tasks to remove\n", 19);
    }
}

void do_modify(int found, char *id, char *name)
{
    char *argv[6];
    argv[0] = "../tm";
    argv[1] = "m";
    argv[2] = "-i";
    argv[3] = id;
    argv[4] = name;
    argv[5] = 0;

    if (1 == found)
    {
        run_with_args(argv, 0, "", 0, "", 0);
    }
    else
    {
        run_with_args(argv, 1, "", 0, "No tasks to modify\n", 19);
    }
}

void do_query_dbfile_v(char *dbfile, char *expected, unsigned int len,
    va_list list)
{
    char **argv;
    unsigned int i = 0;
    unsigned int j = 0;

    argv = (char **)malloc(sizeof(char *) * (5 + len));
    argv[i++] = "../tm";

    if (0 != dbfile)
    {
        argv[i++] = "-d";
        argv[i++] = dbfile;
    }
    argv[i++] = "q";
    for (j = 0; j < len; j++)
    {
        argv[i++] = va_arg(list, char *);
    }
    argv[i] = 0;

    run_with_args(argv, 0, expected, strlen(expected), "", 0);
    free(argv);
}

void do_query(char *expected, unsigned int len, ...)
{
    va_list list;

    va_start(list, len);
    do_query_dbfile_v(0, expected, len, list);
    va_end(list);
}

void do_query_dbfile(char *dbfile, char *expected, unsigned int len, ...)
{
    va_list list;

    va_start(list, len);
    do_query_dbfile_v(dbfile, expected, len, list);
    va_end(list);
}

void init_folder(void)
{
    sprintf(tm_dir, "tm-%d", getpid());
    if (0 != mkdir(tm_dir, 0700))
    {
        cr_expect_fail("Cannot mkdir() %s: %d %s", tm_dir, errno, strerror(errno));
    }
    else if (0 != chdir(tm_dir))
    {
        rmdir(tm_dir);
        cr_expect_fail("Cannot chdir() %s: %d %s", tm_dir, errno, strerror(errno));
    }
}

void clear_folder(void)
{
    remove("tm.db");
    chdir("..");
    rmdir(tm_dir);
}
/* LCOV_EXCL_STOP */
