/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <stdio.h>
#include <string.h>
#include <signal.h>

#include <jansson.h>
#include <mongoose.h>
#include <tm.h>


typedef struct argz argz;
struct argz
{
    int a;
    char *d;
    int h;
    char *i;
    unsigned int p;
};

typedef struct add_modify_args add_modify_args;
struct add_modify_args
{
    char *name;
};

extern char file_static_index_html[];
extern char file_static_bundle_js[];

extern unsigned long file_static_index_html_size;
extern unsigned long file_static_bundle_js_size;

static int should_exit = 0;
static tm *taskman;
static struct mg_mgr *mgr;

static void add_modify_args_free(add_modify_args *amargs)
{
    if (0 != amargs->name)
    {
        free(amargs->name);
    }
}

static void argz_free(argz *argz)
{
    if (0 != argz->d)
    {
        free(argz->d);
    }
    if (0 != argz->i)
    {
        free(argz->i);
    }
}

static void usage(char *argv0, FILE *out)
{
    fprintf(out, "Usage: %s [-h]\n", argv0);
    fprintf(out, "       %s [-a/--access-log] [-d/--dbfile dbfile] "
                 "[-i/--ip ip] [-p/--port port]\n", argv0);
    fprintf(out, "\n");
}

static int parse_opt_key(char *k, char *pl, argz *arg)
{
    if (0 == strcmp("-a", k) ||
        0 == strcmp("--access-log", k))
    {
        arg->a = 1;
    }
    else if (0 == strcmp("-d", k) ||
        0 == strcmp("--dbfile", k))
    {
        *pl = 'd';
    }
    else if (0 == strcmp("-i", k) ||
        0 == strcmp("--ip", k))
    {
        *pl = 'i';
    }
    else if (0 == strcmp("-p", k) ||
        0 == strcmp("--port", k))
    {
        *pl = 'p';
    }
    else
    {
        return 0;
    }
    return 1;
}

static void parse_opt_value(char k, char *v, argz *arg)
{
    if ('d' == k)
    {
        free(arg->d);
        arg->d = strdup(v);
    }
    else if ('i' == k)
    {
        free(arg->i);
        arg->i = strdup(v);
    }
    else if ('p' == k)
    {
        arg->p = atoi(v);
    }
}

static int parse_opts(int argc, char **argv, argz *arg)
{
    char k = 0;
    unsigned int i;
    for (i = 1; i < argc; i++)
    {
        char *argvi = argv[i];
        int key = (k == 0);
        if (1 == key && 0 == parse_opt_key(argvi, &k, arg))
        {
            return 0;
        }
        else if (0 == key)
        {
            parse_opt_value(k, argvi, arg);
            k = 0;
        }
    }
    return 0 != k ? 0 : 1;
}

static int parse_args(int argc, char **argv, argz *arg)
{
    if (argc == 1)
    {
        return 1;
    }

    if (0 == strcmp("-h", argv[1]) || 0 == strcmp("--help", argv[1]))
    {
        arg->h = 1;
        return 1;
    }

    if (0 == parse_opts(argc, argv, arg))
    {
        return 0;
    }

    return 1;
}

static int extract_query(json_t *root, tm_query **pquery)
{
    unsigned int len = 0;
    json_t *id_equals = 0;
    json_t *name_contains = 0;
    tm_query **children = 0;
    unsigned int i = 0;

    if (0 == json_object_size(root))
    {
        return 1;
    }

    if (0 != (id_equals = json_object_get(root, "id-equals")))
    {
        if (0 == json_is_integer(id_equals))
        {
            return 0;
        }
        len++;
    }
    if (0 != (name_contains = json_object_get(root, "name-contains")))
    {
        if (0 == json_is_string(name_contains))
        {
            return 0;
        }
        len++;
    }

    if (0 == len)
    {
        return 0;
    }

    children = (tm_query **)malloc(sizeof(tm_query *) * len);
    memset(children, 0, sizeof(tm_query *) * len);
    if (0 != id_equals)
    {
        tm_query_id_equals(taskman, json_integer_value(id_equals), &(children[i++]));
    }
    if (0 != name_contains)
    {
        tm_query_name_contains(taskman, (char *)json_string_value(name_contains),
            &(children[i++]));
    }

    if (1 == len)
    {
        *pquery = children[0];
        free(children);
    }
    else
    {
        tm_query_and(taskman, len, children, pquery);
    }
    return 1;
}

static void create_query_reply(tm_task **arr, unsigned int arr_len, char **pret)
{
    json_t *ret_json = json_array();
    char *ret_str = 0;

    for (unsigned i = 0; i < arr_len; i++)
    {
        unsigned long id = -1;
        char *name = 0;
        json_t *tmp = 0;
        json_t *entry;

        entry = json_object();

        tm_task_get_id(arr[i], &id);
        tmp = json_integer(id);
        json_object_set(entry, "id", tmp);
        json_decref(tmp);

        tm_task_get_name(arr[i], &name);
        tmp = json_string(name);
        json_object_set(entry, "name", tmp);
        json_decref(tmp);

        json_array_append(ret_json, entry);
        json_decref(entry);
    }

    ret_str = json_dumps(ret_json, JSON_COMPACT);
    json_decref(ret_json);
    *pret = ret_str;
}

static void print_accesslog(char *ip, unsigned int method_len, char *method,
    unsigned int path_len, char *path, unsigned int proto_len, char *proto,
    unsigned int status, unsigned int len)
{

    printf("%s unknown unknown [unknown] \"%.*s %.*s %.*s\" %u %u\n",
        ip, method_len, method, path_len, path, proto_len, proto, status, len);
}

static void send_error(struct mg_connection *conn, struct http_message *hm, int code)
{
    argz *arg = (argz *)conn->mgr->user_data;

    mg_http_send_error(conn, code, "");

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, code, 0);
    }
}

static void send_ok(struct mg_connection *conn, struct http_message *hm,
    char *headers, unsigned int len, char *bytes)
{
    argz *arg = (argz *)conn->mgr->user_data;

    mg_send_head(conn, 200, len, headers);
    if (0 != bytes)
    {
        mg_send(conn, bytes, len);
    }

    if (1 == arg->a)
    {
        char addr[32];
        mg_conn_addr_to_str(conn, addr, sizeof(addr),
            MG_SOCK_STRINGIFY_IP | MG_SOCK_STRINGIFY_REMOTE);
        print_accesslog(addr, (unsigned int) hm->method.len, (char *)hm->method.p,
            (unsigned int) hm->uri.len, (char *) hm->uri.p, (unsigned int) hm->proto.len,
            (char *)hm->proto.p, 200, len);
    }
}

static void do_query(struct mg_connection *conn, struct http_message *hm)
{
    tm_query *query = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    char *reply = 0;
    int ret;

    if (0 == mg_vcmp(&(hm->method), "POST"))
    {
        json_error_t err;
        json_t *root;

        root = json_loadb(hm->body.p, hm->body.len, 0, &err);
        if (0 == root)
        {
            send_error(conn, hm, 400);
            return;
        }

        ret = extract_query(root, &query);
        json_decref(root);
        if (0 == ret)
        {
            send_error(conn, hm, 400);
            return;
        }
    }
    else if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
        return;
    }

    ret = tm_run_query(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */

    create_query_reply(arr, arr_len, &reply);
    tm_task_arr_free(&arr, arr_len);

    send_ok(conn, hm, "Content-Type: application/json", strlen(reply), reply);
    free(reply);
}

static int extract_add(json_t *root, add_modify_args *amargs)
{
    json_t *name = 0;
    if (1 != json_object_size(root) ||
        0 == (name = json_object_get(root, "name")) ||
        0 == json_is_string(name))
    {
       return 0;
    }

    amargs->name = strdup((char *)json_string_value(name));
    return 1;
}

static int extract_modify(json_t *root, tm_query **pquery,
    add_modify_args *pargs)
{
    json_t *query = 0;
    json_t *args = 0;

    if (2 != json_object_size(root))
    {
        return 0;
    }

    if (0 == (query = json_object_get(root, "query")) ||
        0 == json_is_object(query))
    {
        return 0;
    }
    if (0 == (args = json_object_get(root, "args")) ||
        0 == json_is_object(args))
    {
        return 0;
    }

    if (0 == extract_query(query, pquery))
    {
        return 0;
    }

    if (0 == extract_add(args, pargs))
    {
        tm_query_free(pquery);
        return 0;
    }
    return 1;
}

static void do_modify(struct mg_connection *conn, struct http_message *hm)
{
    json_error_t err;
    json_t *root;
    tm_query *query = 0;
    add_modify_args amargs = {0};
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    int ret;

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = extract_modify(root, &query, &amargs);
    json_decref(root);
    if (0 == ret)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_run_query(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */

    if (0 == arr_len)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        if (0 != amargs.name)
        {
            tm_task_arr_set_name(arr, arr_len, amargs.name);
        }
        ret = tm_modify_tasks(taskman, arr, arr_len);

        /* LCOV_EXCL_START */
        if (0 == ret)
        {
            send_error(conn, hm, 500);
        }
        /* LCOV_EXCL_STOP */
        else
        {
            send_ok(conn, hm, 0, 0, 0);
        }
    }

    add_modify_args_free(&amargs);
    tm_task_arr_free(&arr, arr_len);
}

static void do_remove(struct mg_connection *conn, struct http_message *hm)
{
    tm_query *query = 0;
    tm_task **arr = 0;
    unsigned int arr_len = 0;
    int ret;
    json_error_t err;
    json_t *root;

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = extract_query(root, &query);
    json_decref(root);
    if (0 == ret)
    {
        send_error(conn, hm, 400);
        return;
    }

    ret = tm_run_query(taskman, query, &arr, &arr_len);
    tm_query_free(&query);

    /* LCOV_EXCL_START */
    if (0 == ret)
    {
        send_error(conn, hm, 500);
        return;
    }
    /* LCOV_EXCL_STOP */

    if (0 == arr_len)
    {
        send_error(conn, hm, 404);
    }
    else
    {
        ret = tm_remove_tasks(taskman, arr, arr_len);

        /* LCOV_EXCL_START */
        if (0 == ret)
        {
            send_error(conn, hm, 500);
        }
        /* LCOV_EXCL_STOP */
        else
        {
            send_ok(conn, hm, 0, 0, 0);
        }
    }

    tm_task_arr_free(&arr, arr_len);
}

static void do_add(struct mg_connection *conn, struct http_message *hm)
{
    json_error_t err;
    json_t *root;
    add_modify_args amargs = {0};
    tm_task *task = 0;

    if (0 != mg_vcmp(&(hm->method), "POST"))
    {
        send_error(conn, hm, 405);
        return;
    }

    root = json_loadb(hm->body.p, hm->body.len, 0, &err);
    if (0 == root)
    {
        send_error(conn, hm, 400);
        return;
    }

    if (0 == extract_add(root, &amargs))
    {
        send_error(conn, hm, 400);
    }
    else
    {
        tm_task *tasks[1];

        tm_task_new(&task);
        tm_task_set_name(task, amargs.name);
        add_modify_args_free(&amargs);

        tasks[0] = task;
        /* LCOV_EXCL_START */
        if (0 == tm_add_tasks(taskman, tasks, 1))
        {
            send_error(conn, hm, 500);
        }
        /* LCOV_EXCL_STOP */
        else
        {
            send_ok(conn, hm, 0, 0, 0);
        }
        tm_task_free(&task);
    }
    json_decref(root);
}

static void do_ping(struct mg_connection *conn, struct http_message *hm)
{
    if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
    }
    else
    {
        send_ok(conn, hm, 0, 0, 0);
    }
}

static void event_handler(struct mg_connection *conn, int ev, void *ev_data)
{
    struct http_message *hm;

    if (ev != MG_EV_HTTP_REQUEST)
    {
        return;
    }

    hm = (struct http_message *)ev_data;
    if (0 == mg_vcmp(&(hm->uri), "/ping"))
    {
        do_ping(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/query"))
    {
        do_query(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/add"))
    {
        do_add(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/modify"))
    {
        do_modify(conn, hm);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/remove"))
    {
        do_remove(conn, hm);
    }
    else if (0 != mg_vcmp(&(hm->method), "GET"))
    {
        send_error(conn, hm, 405);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/") ||
        0 == mg_vcmp(&(hm->uri), "/index.html"))
    {
        send_ok(conn, hm, "Content-Type: text/html", file_static_index_html_size,
            file_static_index_html);
    }
    else if (0 == mg_vcmp(&(hm->uri), "/bundle.js"))
    {
        send_ok(conn, hm, "Content-Type: application/javascript", file_static_bundle_js_size,
            file_static_bundle_js);
    }
    else
    {
        send_error(conn, hm, 404);
    }
    conn->flags |= MG_F_SEND_AND_CLOSE;
}

static void on_ctrl_c(int signum)
{
    should_exit = 1;
}

static void init_tm(argz *arg)
{
    tm_storage *storage = 0;

    taskman = 0;
    tm_storage_sqlite(arg->d, &storage);
    tm_new(storage, &taskman);
}

static void deinit_tm(void)
{
    tm_free(&taskman);
}

static int init_mgr(argz *arg)
{
    struct mg_connection *conn;
    struct mg_bind_opts bind_opts;
    char addr[22];
    char *error_string = 0;

    mgr = (struct mg_mgr *)malloc(sizeof(struct mg_mgr));
    memset(mgr, 0, sizeof(struct mg_mgr));

    sprintf(addr, "%s:%u", arg->i, arg->p);
    memset(&bind_opts, 0, sizeof(bind_opts));
    bind_opts.error_string = (const char **)&error_string;
    mg_mgr_init(mgr, 0);
    mgr->user_data = arg;

    /* LCOV_EXCL_START */
    if (0 == (conn = mg_bind_opt(mgr, addr, event_handler, bind_opts)))
    {
        fprintf(stderr, "Cannot start server: %s\n",
            error_string);
        return 0;
    }
    /* LCOV_EXCL_STOP */

    mg_set_protocol_http_websocket(conn);
    return 1;
}

static void deinit_mgr(void)
{
    mg_mgr_free(mgr);
    free(mgr);
}

static void loop(void)
{
    printf("Started\n");
    while (0 == should_exit)
    {
         mg_mgr_poll(mgr, 200);
    }
    printf("Stopped\n");
}

int main(int argc, char **argv)
{
    argz arg;
    int ret = 1;

    arg.a = 0;
    arg.d = strdup("tm.db");
    arg.h = 0;
    arg.p = 3805;
    arg.i = strdup("0.0.0.0");

    if (0 == parse_args(argc, argv, &arg))
    {
        usage(argv[0], stderr);
    }
    else if (1 == arg.h)
    {
        usage(argv[0], stdout);
        ret = 0;
    }
    else
    {
        init_tm(&arg);

        if (1 == init_mgr(&arg))
        {
            setvbuf(stdout, 0, _IOLBF, 1024);
            signal(SIGINT, on_ctrl_c);

            loop();

            deinit_mgr();
        }

        deinit_tm();
        ret = 0;
    }
    argz_free(&arg);
    return ret;
}
