/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/


#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>
#include <mongoose.h>

#include "common.h"

/* LCOV_EXCL_START */

struct http_userdata
{
    int should_exit;
    int expected_code;
    char *expected_body;
    int ignore_body;
};

static char *usage = "Usage: ./tms [-h]\n"
                      "       ./tms [-a/--access-log] [-d/--dbfile dbfile] "
                      "[-i/--ip ip] [-p/--port port]\n"
                      "\n";

static pid_t tms_pid;
static int tms_stdout;
static int tms_stderr;
static char tms_dir[16];

static void set_nonblocking(int output)
{
    int opts = fcntl(output, F_GETFL);
    if (0 > opts)
    {
        cr_log_error("Cannot fcntl F_GETFL: %d %s", errno, strerror(errno));
        return;
    }

    opts |= O_NONBLOCK;
    if (0 > fcntl(output, F_SETFL, opts))
    {
        cr_log_error("Cannot fcntl F_SETFL: %d %s", errno, strerror(errno));
    }
}

static int start_process_with_args(char **argv, pid_t *ppid, int *pstdout, int *pstderr)
{
    pid_t pid = 0;
    int stdout[2];
    int stderr[2];
    unsigned int i;

    cr_log_info("# Start process args:");
    for (i = 0; 0 != argv[i]; i++)
    {
        cr_log_info("    argv[%u] = %s", i, argv[i]);
    }

    if (-1 == pipe(stdout))
    {
        cr_expect_fail("Cannot pipe() stdout");
        return 0;
    }
    if (-1 == pipe(stderr))
    {
        cr_expect_fail("Cannot pipe() stderr");
        return 0;
    }

    pid = fork();

    if (pid == -1)
    {
        cr_expect_fail("Cannot fork()");
        return 0;
    }

    if (pid == 0)
    {
        dup2(stdout[1], 1);
        dup2(stderr[1], 2);
        close(stdout[1]);
        close(stderr[1]);
        execv(argv[0], argv);
        cr_expect_fail("Cannot execv()");
        _exit(1);
    }

    close(stdout[1]);
    close(stderr[1]);

    set_nonblocking(stdout[0]);
    set_nonblocking(stderr[0]);

    *ppid = pid;
    *pstdout = stdout[0];
    *pstderr = stderr[0];
    return 1;
}

static void kill_process(pid_t pid, int a_signal)
{
    cr_log_info("# Kill process with signal=%d", a_signal);
    kill(pid, a_signal);
}

static void wait_for_exit_process(pid_t pid, unsigned int expected_exit)
{
    int status = 0;
    int actual_exit;

    cr_log_info("# Wait for exit process");
    cr_log_info("    expected_exit = %d", expected_exit);

    waitpid(pid, &status, 0);
    actual_exit = WEXITSTATUS(status);

    cr_log_info("    actual_exit = %d", actual_exit);
    cr_expect_eq(expected_exit, actual_exit,
        "Exit code different, expected %d, actual %d",
        expected_exit, actual_exit);
}

static int is_read_selected(int stream, unsigned int timeout)
{
    struct timeval second = {timeout, 0};
    fd_set aset;
    int ret;

    FD_ZERO(&aset);
    FD_SET(stream, &aset);

    ret = select(FD_SETSIZE, &aset, 0, 0, &second);
    if (0 > ret)
    {
        cr_expect_fail("Cannot select(): %d %s", errno, strerror(errno));
        return 0;
    }

    return 1;
}

static int is_read_error(int read_ret)
{
    if (0 <= read_ret || EWOULDBLOCK == errno || EAGAIN == errno)
    {
        return 0;
    }
    cr_expect_fail("Cannot read(): %d %s", errno, strerror(errno));
    return 1;
}

static int is_expected_read(char **pexpected, unsigned int *pexpected_len,
    char *buf, unsigned int buf_len)
{
    int ret = 0;
    cr_log_info("    Read %u bytes", buf_len);
    if (buf_len > *pexpected_len)
    {
        cr_expect_fail("Read more bytes than expected. \n"
            "Expected %u bytes\nRead %u: <%.*s>", *pexpected_len, buf_len, buf_len, buf);
    }
    else if (0 != strncmp(buf, *pexpected, buf_len))
    {
        cr_expect_fail("Read invalid data.\n"
            "Expected at most %u: <%.*s>\n"
            "Actually read %u: <%.*s>",
            *pexpected_len, *pexpected_len, *pexpected,
            buf_len, buf_len, buf);
    }
    else
    {
        *pexpected += buf_len;
        *pexpected_len -= buf_len;
        ret = 1;
    }
    return ret;
}

static void expect_output(int stream, char *expected, unsigned int expected_len,
    unsigned int timeout_secs)
{
    char buf[4096];
    cr_log_info("# Expect output len=%u timeout=%u", expected_len, timeout_secs);
    cr_log_info("    expected_len = %u", expected_len);
    cr_log_info("<expected>\n%.*s\n</expected>", expected_len,
        expected);

    do
    {
        int ret;
        if (0 == is_read_selected(stream, timeout_secs))
        {
            break;
        }

        ret = read(stream, buf, expected_len);
        if (1 == is_read_error(ret))
        {
            break;
        }
        else if (0 == ret && 0 < expected_len)
        {
            cr_expect_fail("Read less than expected. "
                "Expected %u more", expected_len);
            break;
        }
        else if (0 == expected_len || 0 == is_expected_read(&expected, &expected_len, buf, ret))
        {
            break;
        }
    }
    while (0 < expected_len);
}

static void skip_output_until_char(int stream, char chr, unsigned int timeout_secs)
{
    char buf[1];
    cr_log_info("# Skip output until char %c (%u)", chr, chr);

    do
    {
        int ret;
        if (0 == is_read_selected(stream, timeout_secs))
        {
            break;
        }

        ret = read(stream, buf, 1);
        if (1 == is_read_error(ret))
        {
            break;
        }
        else if (0 == ret)
        {
            cr_expect_fail("Read less than expected. Expected to encounter %c (%u) finally",
                chr, chr);
            break;
        }
        else if (buf[0] == chr)
        {
            break;
        }
    }
    while (1);
}

static void on_connect(struct http_userdata *puserdata, int status)
{
    if (0 != status)
    {
         cr_expect_fail("Could not connect: %d", status);
         puserdata->should_exit = 1;
    }
}

static void on_reply(struct http_userdata *puserdata, struct http_message *hm)
{
    unsigned int expected_body_len = 0;
    if (0 != puserdata->expected_body)
    {
        expected_body_len = strlen(puserdata->expected_body);
    }

    cr_log_info("    expected code = %d", puserdata->expected_code);
    cr_log_info("    actual_code = %u", hm->resp_code);
    cr_expect_eq(puserdata->expected_code, hm->resp_code,
        "Code different, expected = %u, actual = %u",
        puserdata->expected_code, hm->resp_code);

    if (1 == puserdata->ignore_body)
    {
        return;
    }

    cr_log_info("    <expected>\n%s\n    </expected>", puserdata->expected_body);
    cr_log_info("    <actual>\n%.*s\n    </actual>", (int)hm->body.len, hm->body.p);
    if (0 != puserdata->expected_body && (
        expected_body_len != hm->body.len ||
        0 != strncmp(puserdata->expected_body, hm->body.p, expected_body_len)))
    {
        cr_expect_fail("Reply different, expected = %u, actual = %u\n"
            "<expected>\n%s\n</expected>\n"
            "<actual>\n%.*s\n</actual>",
            expected_body_len, hm->body.len,
            puserdata->expected_body,
            (int) hm->body.len, hm->body.p);
    }
    else if (0 == puserdata->expected_body && 0 != hm->body.len)
    {
        cr_expect_eq(0, hm->body.len,
            "Reply not empty: actual = %u\n"
            "<actual>\n%.*s\n</actual>",
            hm->body.len, hm->body.len, hm->body.p);
    }
}

static void http_event_handler(struct mg_connection *conn, int ev, void *ev_data)
{
    struct http_userdata *puserdata = (struct http_userdata *)conn->mgr->user_data;

    switch (ev)
    {
        case MG_EV_CONNECT:
            on_connect(puserdata, *(int *)ev_data);
            break;

        case MG_EV_HTTP_REPLY:
            on_reply(puserdata, (struct http_message *)ev_data);
            conn->flags |= MG_F_CLOSE_IMMEDIATELY;
            puserdata->should_exit = 1;
            break;

        case MG_EV_CLOSE:
            if (0 == puserdata->should_exit)
            {
                cr_expect_fail("Server closed connection");
                puserdata->should_exit = 1;
            }
            break;
    }
}

void expect_http_get(char *url, int expected_code, char *expected_body)
{
    struct http_userdata userdata;
    struct mg_mgr mgr;

    cr_log_info("# Expect HTTP get");
    cr_log_info("    url = %s", url);

    userdata.should_exit = 0;
    userdata.ignore_body = 0;
    userdata.expected_code = expected_code;
    userdata.expected_body = expected_body;

    mg_mgr_init(&mgr, 0);
    mgr.user_data = &userdata;
    mg_connect_http(&mgr, http_event_handler, url, 0, 0);
    while (0 == userdata.should_exit)
    {
        mg_mgr_poll(&mgr, 250);
    }
    mg_mgr_free(&mgr);
}

void expect_http_get_ignore_body(char *url, int expected_code)
{
    struct http_userdata userdata;
    struct mg_mgr mgr;

    cr_log_info("# Expect HTTP get");
    cr_log_info("    url = %s", url);

    userdata.should_exit = 0;
    userdata.ignore_body = 1;
    userdata.expected_code = expected_code;
    userdata.expected_body = 0;

    mg_mgr_init(&mgr, 0);
    mgr.user_data = &userdata;
    mg_connect_http(&mgr, http_event_handler, url, 0, 0);
    while (0 == userdata.should_exit)
    {
        mg_mgr_poll(&mgr, 250);
    }
    mg_mgr_free(&mgr);
}

void expect_http_post(char *url, char *post_body,
    int expected_code, char *expected_body)
{
    struct http_userdata userdata;
    struct mg_mgr mgr;
    char headers[256];

    cr_log_info("# Expect HTTP post");
    cr_log_info("    url = %s", url);
    cr_log_info("    <post_body>\n%s\n    </post_body>", post_body);

    sprintf(headers, "Content-Length: %lu\r\nContent-Type: application/json\r\n",
        strlen(post_body));

    userdata.should_exit = 0;
    userdata.ignore_body = 0;
    userdata.expected_code = expected_code;
    userdata.expected_body = expected_body;

    mg_mgr_init(&mgr, 0);
    mgr.user_data = &userdata;
    mg_connect_http(&mgr, http_event_handler, url, headers, post_body);
    while (0 == userdata.should_exit)
    {
        mg_mgr_poll(&mgr, 250);
    }
    mg_mgr_free(&mgr);
}

void expect_process_usage(char **argv, int code)
{
    pid_t pid = 0;
    int stdout = 0;
    int stderr = 0;

    char *expected_stdout = "";
    int expected_stdout_len = 0;
    char *expected_stderr = "";
    int expected_stderr_len = 0;

    if (0 == code)
    {
        expected_stdout = usage;
        expected_stdout_len = strlen(usage);
    }
    else
    {
        expected_stderr = usage;
        expected_stderr_len = strlen(usage);
    }

    if (0 == start_process_with_args(argv, &pid, &stdout, &stderr))
    {
        return;
    }
    wait_for_exit_process(pid, code);
    expect_output(stdout, expected_stdout, expected_stdout_len, 1);
    expect_output(stderr, expected_stderr, expected_stderr_len, 1);
}

void expect_process_started_and_died(char **argv)
{
    pid_t pid = 0;
    int stdout = 0;
    int stderr = 0;

    if (0 == start_process_with_args(argv, &pid, &stdout, &stderr))
    {
        return;
    }
    expect_output(stdout, "Started\n", 8, 2);
    expect_http_get("http://127.0.0.1:3805/ping", 200, "");
    kill_process(pid, SIGINT);
    wait_for_exit_process(pid, 0);
    expect_output(stdout, "Stopped\n", 8, 0);
    expect_output(stderr, "", 0, 0);
}

static void expect_process_started2(int access_log)
{
    char *argv[4];
    unsigned int i = 0;

    sprintf(tms_dir, "tm-%d", getpid());
    if (0 != mkdir(tms_dir, 0700))
    {
        cr_expect_fail("Cannot mkdir() %s: %d %s", tms_dir, errno, strerror(errno));
        return;
    }
    if (0 != chdir(tms_dir))
    {
        rmdir(tms_dir);
        cr_expect_fail("Cannot chdir() %s: %d %s", tms_dir, errno, strerror(errno));
        return;
    }

    argv[i++] = "../tms";
    if (1 == access_log)
    {
        argv[i++] = "-a";
    }
    argv[i++] = 0;

    if (0 == start_process_with_args(argv, &tms_pid, &tms_stdout, &tms_stderr))
    {
        return;
    }
    expect_output(tms_stdout, "Started\n", 8, 2);
}

void expect_process_started(void)
{
    expect_process_started2(0);
}

void expect_process_started_accesslog(void)
{
    expect_process_started2(1);
}

void expect_process_no_output(void)
{
    expect_output(tms_stdout, "", 0, 0);
    expect_output(tms_stderr, "", 0, 0);
}

void expect_process_output_accesslog(char *ip, char *method, char *url, unsigned int code,
    unsigned int length)
{
    char tmp[32];

    expect_output(tms_stdout, ip, strlen(ip), 0);
    expect_output(tms_stdout, " ", 1, 0);
    expect_output(tms_stdout, "unknown", 7, 0);
    expect_output(tms_stdout, " ", 1, 0);
    expect_output(tms_stdout, "unknown", 7, 0);
    expect_output(tms_stdout, " [", 2, 0);
    /* 10/Oct/2000:13:55:36 -0700 */
    skip_output_until_char(tms_stdout, ']', 0);
    expect_output(tms_stdout, " \"", 2, 0);
    expect_output(tms_stdout, method, strlen(method), 0);
    expect_output(tms_stdout, " ", 1, 0);
    expect_output(tms_stdout, url, strlen(url), 0);
    expect_output(tms_stdout, " ", 1, 0);
    expect_output(tms_stdout, "HTTP/1.1\" ", 10, 0);

    sprintf(tmp, "%u", code);
    expect_output(tms_stdout, tmp, strlen(tmp), 0);
    expect_output(tms_stdout, " ", 1, 0);

    sprintf(tmp, "%u", length);
    expect_output(tms_stdout, tmp, strlen(tmp), 0);
    expect_output(tms_stdout, "\n", 1, 0);
}

void expect_process_died(void)
{
    kill_process(tms_pid, SIGINT);
    wait_for_exit_process(tms_pid, 0);
    expect_output(tms_stdout, "Stopped\n", 8, 0);
    expect_output(tms_stderr, "", 0, 0);
    remove("tm.db");
    chdir("..");
    rmdir(tms_dir);
}

void do_add(char *name)
{
    json_t *request_json = json_object();
    char *request_str = 0;
    json_t *tmp = 0;

    tmp = json_string(name);
    json_object_set(request_json, "name", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);
    expect_http_post("http://127.0.0.1:3805/add", request_str, 200, 0);
    free(request_str);

    expect_process_no_output();
}

static void query_args_to_json(json_t *query, va_list list)
{
    char *key = va_arg(list, char *);
    json_t *value;

    if (0 == strcmp(key, "id-equals"))
    {
        value = json_integer(va_arg(list, unsigned int));
    }
    else if (0 == strcmp(key, "name-contains"))
    {
        value = json_string(va_arg(list, char *));
    }
    json_object_set(query, key, value);
    json_decref(value);
}

void do_remove(int found, unsigned int args_len, ...)
{
    char *request_str = 0;
    va_list list;

    va_start(list, args_len);
    json_t *request_json = json_object();
    if (0 != args_len)
    {
        unsigned int i;
        for (i = 0; i < args_len; i++)
        {
            query_args_to_json(request_json, list);
        }
        request_str = json_dumps(request_json, JSON_COMPACT);
        json_decref(request_json);
    }
    va_end(list);

    expect_http_post("http://127.0.0.1:3805/remove", request_str,
        1 == found ? 200 : 404, 0);
    free(request_str);

    expect_process_no_output();
}

void do_modify(int found, unsigned int query_len, unsigned int args_len, ...)
{
    json_t *request_json = json_object();
    json_t *query = json_object();
    json_t *args = json_object();
    char *request_str = 0;
    va_list list;
    int i;

    va_start(list, args_len);
    json_object_set(request_json, "query", query);
    for (i = 0; i < query_len; i++)
    {
        query_args_to_json(query, list);
    }
    json_decref(query);

    json_object_set(request_json, "args", args);
    for (i = 0; i < args_len; i++)
    {
        char *key = va_arg(list, char *);
        json_t *value = json_string(va_arg(list, char *));
        json_object_set(args, key, value);
        json_decref(value);
    }
    json_decref(args);
    va_end(list);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_http_post("http://127.0.0.1:3805/modify", request_str,
        1 == found ? 200 : 404, 0);
    free(request_str);

    expect_process_no_output();
}

void do_query(unsigned int expected_len, unsigned int args_len, ...)
{
    json_t *expected_ret_json = json_array();
    char *expected_ret_str = 0;
    char *request_str = 0;
    unsigned int i;
    va_list list;

    va_start(list, args_len);
    for (i = 0; i < expected_len; i++)
    {
        query_row *row = va_arg(list, query_row *);
        json_t *item = json_object();
        json_t *tmp;

        tmp = json_integer(row->id);
        json_object_set(item, "id", tmp);
        json_decref(tmp);

        tmp = json_string(row->name);
        json_object_set(item, "name", tmp);
        json_decref(tmp);

        json_array_append(expected_ret_json, item);
        json_decref(item);
    }
    if (0 != args_len)
    {
        json_t *request_json = json_object();
        for (i = 0; i < args_len; i++)
        {
            query_args_to_json(request_json, list);
        }
        request_str = json_dumps(request_json, JSON_COMPACT);
        json_decref(request_json);
    }
    va_end(list);

    expected_ret_str = json_dumps(expected_ret_json, JSON_COMPACT);
    json_decref(expected_ret_json);
    if (0 != request_str)
    {
        expect_http_post("http://127.0.0.1:3805/query", request_str, 200, expected_ret_str);
        free(request_str);
    }
    else
    {
        expect_http_get("http://127.0.0.1:3805/query", 200, expected_ret_str);
    }
    free(expected_ret_str);

    expect_process_no_output();
}

/* LCOV_EXCL_STOP */
