/*
 * Copyright (c) 2018 Tomasz Sieprawski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 **/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <criterion/criterion.h>
#include <jansson.h>

#include "common.h"

TestSuite(modify);

Test(modify, id_equals)
{
    query_row beer = {1, "Buy beer"};

    expect_process_started();
    do_add("Buy milk");
    do_modify(1, 1, 1,
        "id-equals", 1,
        "name", "Buy beer");
    do_query(1, 0,
        &beer);
    expect_process_died();
}

Test(modify, name_contains)
{
    query_row beer = {1, "Buy beer"};

    expect_process_started();
    do_add("Buy milk");
    do_modify(1, 1, 1,
        "name-contains", "milk",
        "name", "Buy beer");
    do_query(1, 0,
        &beer);
    expect_process_died();
}

Test(modify, invalid_not_json)
{
    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", "}", 400, 0);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_not_2_items)
{
    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", "{}", 400, 0);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_no_query)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = json_object();
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_query_not_object)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp;

    tmp = json_integer(0);
    json_object_set(request_json, "query", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_no_args)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = json_object();
    json_object_set(request_json, "query", tmp);
    json_decref(tmp);
    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_args_not_object)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp;

    tmp = json_integer(0);
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "query", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_args_empty)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;

    tmp = json_object();
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);

    tmp = json_object();
    json_object_set(request_json, "query", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_query_id_equals_not_int)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;

    tmp = json_object();
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_string("invalid");
    json_object_set(tmp, "id-equals", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "query", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_query_name_contains_not_string)
{
    char *request_str = 0;
    json_t *request_json = json_object();
    json_t *tmp = 0;
    json_t *tmp2 = 0;

    tmp = json_object();
    json_object_set(request_json, "args", tmp);
    json_decref(tmp);

    tmp = json_object();
    tmp2 = json_integer(0);
    json_object_set(tmp, "name-contains", tmp2);
    json_decref(tmp2);

    json_object_set(request_json, "query", tmp);
    json_decref(tmp);

    request_str = json_dumps(request_json, JSON_COMPACT);
    json_decref(request_json);

    expect_process_started();
    expect_http_post("http://127.0.0.1:3805/modify", request_str, 400, 0);
    free(request_str);
    expect_process_no_output();
    expect_process_died();
}

Test(modify, invalid_not_post)
{
    expect_process_started();
    expect_http_get("http://127.0.0.1:3805/modify", 405, 0);
    expect_process_no_output();
    expect_process_died();
}
